# -*- coding: utf-8 -*-

import reg
from uritools import urisplit


TRUEVALUES = (u'true',)


def url2dict(url, raise_on_extra_params=False):
    """

    :param url: Access-URL, see below for supported forms
    :return: Parameter dictionary suitable for get_store()

    memory://
    redis://[[password@]host[:port]][/db]
    fs://path
    s3://access_key:secret_key@endpoint/bucket[?create_if_missing=true]
    azure://account_name:account_key@container[?create_if_missing=true]
    azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false
    """
    u = urisplit(url)
    parsed = dict(
        scheme=u.getscheme(),
        host=u.gethost(),
        port=u.getport(),
        path=u.getpath(),
        query=u.getquerydict(),
        userinfo=u.getuserinfo(),
    )
    fragment = u.getfragment()

    params = {'type': parsed['scheme']}

    # handling special instructions embedded in the 'fragment' part of the URL,
    # currently only wrappers/store decorators
    fragments = fragment.split('#') if fragment else []
    wrap_spec = list(filter(lambda s: s.startswith('wrap:'), fragments))
    if wrap_spec:
        wrappers = wrap_spec[-1].partition('wrap:')[2]  # remove the 'wrap:' part
        params['wrap'] = wrappers

    if u'create_if_missing' in parsed['query']:
        create_if_missing = parsed['query'].pop(u'create_if_missing')[-1]  # use last appearance of key
        params['create_if_missing'] = create_if_missing in TRUEVALUES

    # get store-specific parameters:
    store_params = extract_params(**parsed)
    params.update(store_params)
    return params


@reg.dispatch(reg.match_key('scheme', lambda scheme, host, port, path, query, userinfo: scheme))
def extract_params(scheme, host, port, path, query, userinfo):
    raise ValueError('Unknown storage type "{}"'.format(scheme))


@extract_params.register(scheme='hmemory')
@extract_params.register(scheme='memory')
def extract_params_memory(scheme, host, port, path, query, userinfo):
    return {}


@extract_params.register(scheme='hredis')
@extract_params.register(scheme='redis')
def extract_params_redis(scheme, host, port, path, query, userinfo):
    path = path[1:] if path.startswith(u'/') else path
    params = {'host': host or u'localhost'}
    if port:
        params['port'] = port
    if userinfo:
        params['password'] = userinfo
    if path:
        params['db'] = int(path)
    return params


@extract_params.register(scheme='fs')
@extract_params.register(scheme='hfs')
def extract_params_fs(scheme, host, port, path, query, userinfo):
    return {'type': scheme, 'path': host + path}


@extract_params.register(scheme='s3')
@extract_params.register(scheme='hs3')
def extract_params_s3(scheme, host, port, path, query, userinfo):
    access_key, secret_key = _parse_userinfo(userinfo)
    params = {
        'host': u'{}:{}'.format(host, port) if port else host,
        'access_key': access_key,
        'secret_key': secret_key,
        'bucket': path[1:],
    }
    return params


@extract_params.register(scheme='hazure')
@extract_params.register(scheme='azure')
def extract_params_azure(scheme, host, port, path, query, userinfo):
    account_name, account_key = _parse_userinfo(userinfo)
    params = {
        'account_name': account_name,
        'account_key': account_key,
        'container': host,
    }
    if u'use_sas' in query:
        params['use_sas'] = True
    return params


def _parse_userinfo(userinfo):
    """Try to split the URL's userinfo (the part between :// and @) into fields
    separated by :. If anything looks wrong, remind user to percent-encode values."""
    if hasattr(userinfo, 'split'):
        parts = userinfo.split(u':', 1)

        if len(parts) == 2:
            return parts

    raise ValueError('Could not parse user/key in store-URL. Note that values have to be percent-encoded, eg. with urllib.quote_plus().')




# -*- coding: utf-8 -*-
""""""

from __future__ import (absolute_import, division, print_function)


def _get_s3bucket(host, bucket, access_key, secret_key, force_bucket_suffix=True, create_if_missing=True):
    from boto.s3.connection import S3Connection, OrdinaryCallingFormat, S3ResponseError

    s3con = S3Connection(aws_access_key_id=access_key,
                         aws_secret_access_key=secret_key,
                         host=host, is_secure=False,
                         calling_format=OrdinaryCallingFormat())
    # add access key prefix to bucket name, unless explicitly prohibited
    if force_bucket_suffix and not bucket.lower().endswith('-' + access_key.lower()):
        bucket = bucket + '-' + access_key.lower()
    try:
        return s3con.get_bucket(bucket)
    except S3ResponseError as ex:
        if ex.status == 404:
            if create_if_missing:
                return s3con.create_bucket(bucket)
            else:
                raise IOError("Bucket {} does not exist".format(bucket))
        raise



# -*- coding: utf-8 -*-

from setuptools_scm import get_version

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.extlinks',
    'reno.sphinxext',
]

releases_debug = True
releases_release_uri = 'https://github.com/blue-yonder/storefact/releases/tag/%s'

extlinks = {'issue': ('https://github.com/blue-yonder/storefact/issues/%s', 'issue')}

source_suffix = '.rst'
master_doc = 'index'
project = u'Store factory for simplekv'
copyright = u'2015-2017 Blue Yonder GmbH'
version = get_version(root='..')

import sphinx_rtd_theme
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

pygments_style = 'trac'
templates_path = ['.']
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = True
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)
html_theme_options = {
}



# -*- coding: utf-8 -*-
""""""

from __future__ import (absolute_import, division, print_function)

import reg
from simplekv.decorator import (URLEncodeKeysDecorator, ReadOnlyDecorator)


@reg.dispatch(reg.match_key('decoratorname', lambda decoratorname, store: decoratorname.split('(')[0]))
def decorate_store(store, decoratorname):
    raise ValueError('Unknown store decorator: ' + str(decoratorname))


@decorate_store.register(decoratorname='urlencode')
def _urlencode(store, decoratorname):
    return URLEncodeKeysDecorator(store)


@decorate_store.register(decoratorname='readonly')
def _readonly(store, decoratorname):
    return ReadOnlyDecorator(store)



# -*- coding: utf-8 -*-
""""""
from __future__ import (absolute_import, division, print_function)

from ._compat import (PY2, reduce)
from ._store_creation import create_store
from ._store_decoration import decorate_store
from ._urls import url2dict
from ._version import version as __version__

__all__ = [
    'get_store', 'get_store_from_url', '__version__',
]


def get_store_from_url(url):
    """
    Take a URL and return a simplekv store according to the parameters in the URL.

    .. note::
       User credentials like secret keys have to be percent-encoded before they can be
       used in a URL (see *azure* and *s3* store types), since they can contain characters
       that are not valid in this part of a URL, like forward-slashes.


       You can use Python to percent-encode your secret key on the commandline like so::

           $ python -c "import urllib; print urllib.quote_plus('''dead/beef''')"
           dead%2Fbeef

    :param url: Access-URL, see below for supported forms
    :return: Parameter dictionary suitable for get_store()

    Store types and URL forms:

    * DictStore: ``memory://``
    * RedisStore: ``redis://[[password@]host[:port]][/db]``
    * FilesystemStore: ``fs://path``
    * BotoStore ``s3://access_key:secret_key@endpoint/bucket[?create_if_missing=true]``
    * AzureBlockBlockStorage: ``azure://account_name:account_key@container[?create_if_missing=true]``
    * AzureBlockBlockStorage (SAS): ``azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false``
    """
    return get_store(**url2dict(url))


def get_store(type, create_if_missing=True, **params):
    """Return a storage object according to the `type` and additional parameters.

    The *type* must be one of the types below, where each allows
    different parameters:

    * ``"azure"``: Returns a ``simplekv.azure.AzureBlockBlobStorage``. Parameters are
      ``"account_name"``, ``"account_key"``, ``"container"``, ``"use_sas"`` and ``"create_if_missing"`` (default: ``True``).
      ``"create_if_missing"`` has to be ``False`` if ``"use_sas"`` is set. When ``"use_sas"`` is set,
      ``"account_key"`` is interpreted as Shared Access Signature (SAS) token.FIRE
    * ``"s3"``: Returns a plain ``simplekv.net.botostore.BotoStore``.
      Parameters must include ``"host"``, ``"bucket"``, ``"access_key"``, ``"secret_key"``.
      Optional parameters are

       - ``"force_bucket_suffix"`` (default: ``True``). If set, it is ensured that
         the bucket name ends with ``-<access_key>``
         by appending this string if necessary;
         If ``False``, the bucket name is used as-is.
       - ``"create_if_missing"`` (default: ``True`` ). If set, creates the bucket if it does not exist;
         otherwise, try to retrieve the bucket and fail with an ``IOError``.
    * ``"hs3"`` returns a variant of ``simplekv.net.botostore.BotoStore`` that allows "/" in the key name.
      The parameters are the same as for ``"s3"``
    * ``"fs"``: Returns a ``simplekv.fs.FilesystemStore``. Specify the base path as "path" parameter.
    * ``"hfs"`` returns a variant of ``simplekv.fs.FilesystemStore``  that allows "/" in the key name.
      The parameters are the same as for ``"file"``.
    * ``"memory"``: Returns a DictStore. Doesn't take any parameters
    * ``"redis"``: Returns a RedisStore. Constructs a StrictRedis using params as kwargs.
      See StrictRedis documentation for details.

    :param str type: Type of storage to open, with optional storage decorators
    :param boolean create_if_missing: Create the "root" of the storage (Azure container, parent directory, S3 bucket, etc.).
      Has no effect for stores where this makes no sense, like `redis` or `memory`.
    :param kwargs: Parameters specific to the Store-class"""

    # split off old-style wrappers, if any:
    parts = type.split('+')
    type = parts.pop(-1)
    decorators = list(reversed(parts))

    # find new-style wrappers, if any:
    wrapspec = params.pop('wrap', '')
    wrappers = list(wrapspec.split('+')) if wrapspec else []

    # can't have both:
    if wrappers:
        if decorators:
            raise ValueError('Adding store wrappers via store type as well as via wrap parameter are not allowed. Preferably use wrap.')
        decorators = wrappers

    # create_if_missing is a universal parameter, so it's part of the function signature
    # it can be safely ignored by stores where 'creating' makes no sense.
    params['create_if_missing'] = create_if_missing

    store = create_store(type, params)

    # apply wrappers/decorators:
    wrapped_store = reduce(decorate_store, decorators, store)

    return wrapped_store



# -*- coding: utf-8 -*-
""""""

from __future__ import (absolute_import, division, print_function)
import os
try:
    import copy_reg as copyreg
except ImportError:
    import copyreg

from simplekv.contrib import ExtendedKeyspaceMixin
from simplekv.fs import FilesystemStore
from simplekv.memory import DictStore
from simplekv.memory.redisstore import RedisStore
from simplekv.net.azurestore import AzureBlockBlobStore, pickle_azure_store
from simplekv.net.botostore import BotoStore


class HDictStore(ExtendedKeyspaceMixin, DictStore):
    pass


class HRedisStore(ExtendedKeyspaceMixin, RedisStore):
    pass


class HAzureBlockBlobStore(ExtendedKeyspaceMixin, AzureBlockBlobStore):
    pass


class HBotoStore(ExtendedKeyspaceMixin, BotoStore):
    def size(self, key):
        k = self.bucket.lookup(self.prefix + key)
        return k.size


class HFilesystemStore(ExtendedKeyspaceMixin, FilesystemStore):
    def size(self, key):
        return os.path.getsize(self._build_filename(key))

copyreg.pickle(HAzureBlockBlobStore, pickle_azure_store)



# -*- coding: utf-8 -*-
""""""

from __future__ import (absolute_import, division, print_function)

import os
import os.path

from simplekv.fs import FilesystemStore
import reg


@reg.dispatch(reg.match_key('type'))
def create_store(type, params):
    raise ValueError('Unknown store type: ' + str(type))


@create_store.register(type='hazure')
@create_store.register(type='azure')
def _create_store_azure(type, params):
    from simplekv.net.azurestore import AzureBlockBlobStore
    from ._hstores import HAzureBlockBlobStore

    conn_string = params.get('connection_string', _build_azure_url(**params))
    
    if params['create_if_missing'] and params.get('use_sas', False):
        raise Exception('create_if_missing is incompatible with the use of SAS tokens.')

    if type == 'azure':
        return AzureBlockBlobStore(conn_string=conn_string, container=params['container'], public=False, create_if_missing=params['create_if_missing'])
    else:
        return HAzureBlockBlobStore(conn_string=conn_string, container=params['container'], public=False, create_if_missing=params['create_if_missing'])


@create_store.register(type='hs3')
@create_store.register(type='boto')
def _create_store_hs3(type, params):
    from ._boto import _get_s3bucket
    from ._hstores import HBotoStore
    return HBotoStore(_get_s3bucket(**params))


@create_store.register(type='s3')
def _create_store_s3(type, params):
    from simplekv.net.botostore import BotoStore
    from ._boto import _get_s3bucket
    return BotoStore(_get_s3bucket(**params))


@create_store.register(type='hfs')
@create_store.register(type='hfile')
@create_store.register(type='filesystem')
def _create_store_hfs(type, params):
    if params['create_if_missing'] and not os.path.exists(params['path']):
        os.makedirs(params['path'])
    from ._hstores import HFilesystemStore
    return HFilesystemStore(params['path'])


@create_store.register(type='fs')
@create_store.register(type='file')
def _create_store_fs(type, params):
    if params['create_if_missing'] and not os.path.exists(params['path']):
        os.makedirs(params['path'])
    return FilesystemStore(params['path'])


@create_store.register(type='memory')
def _create_store_mem(type, params):
    from simplekv.memory import DictStore
    return DictStore()


@create_store.register(type='hmemory')
def _create_store_mem(type, params):
    from ._hstores import HDictStore
    return HDictStore()


@create_store.register(type='redis')
def _create_store_redis(type, params):
    from simplekv.memory.redisstore import RedisStore
    from redis import StrictRedis
    r = StrictRedis(**params)
    return RedisStore(r)


def _build_azure_url(
    account_name=None, account_key=None, default_endpoints_protocol=None, blob_endpoint=None,
        use_sas=False, **kwargs):
    protocol = default_endpoints_protocol or 'https'
    if use_sas:
        return ('DefaultEndpointsProtocol={protocol};AccountName={account_name};'
                'SharedAccessSignature={shared_access_signature}'.format(
                    protocol=protocol, account_name=account_name, shared_access_signature=account_key))
    else:
        return 'DefaultEndpointsProtocol={protocol};AccountName={account_name};AccountKey={account_key}'.format(
            protocol=protocol, account_name=account_name, account_key=account_key)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
import sys

PY2 = sys.version_info[0] == 2
if PY2:
    text_type = unicode
    string_types = (str, unicode)
    unichr = unichr
    binary_type = str
    from itertools import imap
    reduce = reduce
    xrange = xrange
    from ConfigParser import ConfigParser
    from cStringIO import StringIO as BytesIO
    import cPickle as pickle
    import copy_reg as copyreg
else:
    text_type = str
    string_types = (str,)
    unichr = chr
    binary_type = bytes
    imap = map
    from functools import reduce
    xrange = range
    from configparser import ConfigParser
    from io import BytesIO
    import pickle
    import copyreg



# -*- coding: utf-8 -*-

import pytest
import storefact
import simplekv.decorator

good_urls = [
    (u'azure://MYACCOUNT:dead%2Fbeef@1buc-ket1?param1=foo&param2=🍺&eat_more_🍎=1&create_if_missing=true', dict(
        type='azure',
        account_name='MYACCOUNT',
        account_key='dead/beef',
        container='1buc-ket1',
        create_if_missing=True)),
    (u'azure://MYACCOUNT:deadbeef@1bucket1?param1=foo&param2=🍺&eat_more_🍎=1&create_if_missing=true', dict(
        type='azure',
        account_name='MYACCOUNT',
        account_key='deadbeef',
        container='1bucket1',
        create_if_missing=True)),
    (u'fs://this/is/a/relative/path', dict(type='fs', path='this/is/a/relative/path')),
    (u'fs:///an/absolute/path', dict(type=u'fs', path=u'/an/absolute/path')),
    (u's3://access_key:secret_key@endpoint:1234/bucketname', dict(
        type=u's3',
        host=u'endpoint:1234',
        access_key=u'access_key',
        secret_key=u'secret_key',
        bucket=u'bucketname')),
    (u'redis:///2', dict(type=u'redis', host=u'localhost', db=2)),
    (u'memory://#wrap:readonly', {'type':u'memory', 'wrap': u'readonly'}),
    (u'memory://', dict(type=u'memory')),
]

bad_urls = [
    (u'azure://MYACCOUNT:deadb/eef@1buc-ket1?param1=foo&param2=🍺&eat_more_🍎=1&create_if_missing=true',
     ValueError,
    ),
]


def test_raise_on_invalid_store():
    with pytest.raises(ValueError):
        storefact.url2dict(u'dummy://foo/bar')


@pytest.mark.parametrize('url, expected', good_urls)
def test_url2dict(url, expected):
    assert storefact.url2dict(url) == expected


@pytest.mark.parametrize('url, raises', bad_urls)
def test_bad_url2dict(url, raises):
    with pytest.raises(raises):
        storefact.url2dict(url)


def test_roundtrip():
    assert isinstance(storefact.get_store_from_url(u'memory://#wrap:readonly'), simplekv.decorator.ReadOnlyDecorator)



# -*- encoding: utf-8 -*-
import io
import re
import os
from os.path import dirname
from os.path import join
import inspect

from setuptools import setup


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ).read()

instreq = [l.rstrip() for l in read('requirements.txt').splitlines()
           if not l.startswith('#')]


setup(
    name="storefact",
    use_scm_version=dict(write_to='storefact/_version.py'),
    description="A factory for simplekv-Store-based storage classes. Takes configuration values and returns a simplekv-Store",
    author="Felix Marczinowski",
    author_email="felix.marczinowski@blue-yonder.com",
    url="https://github.com/blue-yonder/storefact",
    packages=['storefact'],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Utilities",
    ],
    install_requires=instreq,
    extras_require={
        's3': ['boto'],
        'azure': ['azure-storage-blob'],
    },
    setup_requires=['setuptools_scm']
)
